import random 
import events

class Piece(object):
    """This class represents a Tetris piece, which takes one of seven defined forms."""

    class Type(object):
        """This inner class represents a Tetris piece's type, an enumeration seven basic types."""

        def __init__(self,
                     name,
                     rotation_factor,
                     body):
            object.__init__(self)
            self._name = name
            self._rotation_factor = rotation_factor
            self._body = body

        @property
        def name(self):
            """The name of this piece type."""
            return self._name

        @property
        def rotation_factor(self):
            """The rotation factor of this piece type.  Takes the value of either 0, 2 or 4."""
            return self._rotation_factor

        @property
        def body(self):
            """The (relative) coordinates of the four blocks that make up this piece type, represented as a 4-tuple of 2-tuple (x,y) coordinate points.  The point (0,0) signifies the center block of the tetris piece."""
            return self._body

        def __str__(self):
            return self._name


    # refactor: decouple? XXX
    types = {
                'L': Type('L', 4, ((0,0), (0,1),  (1,0),  (2,0))),
                'J': Type('J', 4, ((0,0), (-1,0), (-2,0), (0,1))),
                'S': Type('S', 2, ((0,0), (1,0),  (0,1),  (-1,1))),
                'O': Type('O', 0, ((0,0), (1,0),  (0,1),  (1,1))),
                'I': Type('I', 2, ((0,0), (-1,0), (1,0),  (2,0))),
                'Z': Type('Z', 2, ((0,0), (-1,0), (0,1),  (1,1))),
                'T': Type('T', 4, ((0,0), (-1,0), (1,0),  (0,1))),
            }


    def __init__(self, 
                 boardref, 
                 type,
                 center=(0,0)):
        object.__init__(self)
        self._boardref = boardref
        self._type = type
        self._body = type.body
        self._center = center
        self._rotation = 0

    @property
    def boardref(self):
        """Get the reference to the Board this piece belongs to."""
        return self._boardref

    @property
    def type(self):
        """Get the Type of this piece."""
        return self._type

    def _get_body(self):
        return self._body
    def _set_body(self, nb):
        if nb and len(nb) == 4:
            self._body = nb
        else:
            raise ValueError
    body = property(_get_body, _set_body, doc='The relative coordinates of the four blocks that make up this piece, represented as a 4-tuple of 2-tuple (x,y) coordinates.')

    def _get_center(self):
        return self._center
    def _set_center(self, nc):
        if nc and len(nc) == 2:
            self._center = nc
        else:
            raise ValueError
    center = property(_get_center, _set_center, doc="The real coordinate of the piece's central block on the board.")

    def move(self, direction):
        """Move this piece one unit in the direction provided.  Returns a boolean value representing whether the new proposed position for the piece is valid. If not, the action changes nothing."""
        (x, y) = self._center
        bref = self._boardref

        def __default_backtrack():
            self._center = (x, y)

        def __mover(forward, backtrack):
            result = True

            bref.remove_piece(self)
            forward()

            if not bref.add_piece(self, True):
                backtrack()
                bref.add_piece(self, True)
                result = False

            return result
            
        if direction == 'FALL':
            loop = True
            result = True
            while loop:
                (x, y) = self._center
                bref.remove_piece(self)
                self._center = (x, y+1)
                
                if not bref.add_piece(self, False):
                    self._center = (x, y)
                    bref.add_piece(self, True)
                    loop = False
                    result = False
        else:
            backtrack = __default_backtrack

            if direction == 'LEFT':
                def forward():
                    self._center = (x-1, y)
            elif direction == 'RIGHT':
                def forward():
                    self._center = (x+1, y)
            elif direction == 'DOWN':
                def forward():
                    self._center = (x, y+1)
            elif direction == 'ROTATE':
                def forward():
                    self._rotate_clockwise()
                def backtrack():
                    self._rotate_clockwise()
                    self._rotate_clockwise()
                    self._rotate_clockwise()
            else: 
                raise ValueError
            result = __mover(forward, backtrack)

        return result

    def _rotate_clockwise(self):
        """Rotates the piece clockwise along the center block of the piece."""
        rotf = self._type.rotation_factor

        if rotf > 0:
            self._rotation = self._rotation + 1
            if rotf == 2 and self._rotation == 2:
                self._do_rotate()
                self._do_rotate()
                self._do_rotate()
            else:
                self._do_rotate()
            self._rotation = self._rotation % rotf

    def _do_rotate(self):
        """Apply the actual rotation algorithm on each of the block."""
        nbod = [(-y, x) for (x, y) in self._body]
        self._body = tuple(nbod)

    @staticmethod
    def get_random_piece(board):
        """Returns a tetris piece that belongs to the provided board reference. The type of the piece is chosen at random."""
        pick = random.choice(Piece.types.keys())
        type = Piece.types[pick]
        return Piece(board, type)

    def __str__(self):
        return self._type.__str__() + '-piece at ' + self._center.__str__() + '; rotf=' + self._rotation.__str__() + '; body=' + self._body.__str__()
       

class Board(object):
    """This class represents a Tetris board, which is an m by n cell grid of Tetris piece types."""

    def __init__(self, width, height):
        object.__init__(self)
        self._width = width
        self._height = height
        self._listeners = []
        self._reset_board()

    def _reset_board(self):
        self._cells = dict([((x,y), None) for x in range(self._width) for y in range(self._height)])

    def _get_width(self):
        return self._width
    def _set_width(self, nw):
        self._width = nw
        self._reset_board()
    width = property(_get_width, _set_width, doc='The number of cells this board has accrpss.')

    def _get_height(self):
        return self._height
    def _set_height(self, nh):
        self._height = nh
        self._reset_board()
    height = property(_get_height, _set_height, doc='The number of cells this board has down.')

    def set_dimension(self, width, height):
        """Sets the width and height of this board to new values nw and nh, then resets the board."""
        self._width = width
        self._height = height
        self._reset_board()

    def get_piece_type_at(self, x, y):
        """Gets the Type that occupies a cell at coordinate (x,y) where (0,0) is the top left hand corner of the board cells."""
        return self._cells[(x,y)]

    def add_piece(self, piece, notify):
        """See if the piece can be added to the board. Do so and return True if it's valid.  Otherwise do nothing and return False."""
        if (piece == None or not self._will_fit(piece)):
            return False

        (cx,cy) = piece.center
        for (x,y) in piece.body:
            self._cells[(cx+x,cy+y)] = piece.type.name

        if notify:
            self._fire_board_event()

        return True

    def _will_fit(self, piece):
        (cx,cy) = piece.center
        for (x,y) in piece.body:
            try:
                if self._cells[(cx+x,cy+y)] != None:
                    return False
            except KeyError:
                return False
        return True


    def remove_piece(self, piece):
        """Removes the said piece from the board."""
        if piece == None:
            return
        if piece.boardref != self:
            raise AssertionErroe

        (cx,cy) = piece.center
        for (x,y) in piece.body:
            self._cells[(cx+x,cy+y)] = None

    def remove_row(self, row):
        """Remove all cells in the row r by bringing all the cells above said row one unit down, then add an all-None row at the top."""
        r = range(1, row+1)
        # shift down all rows above the removed row
        for i in reversed(r):
            for j in range(self._width):
                self._cells[(j,i)] = self._cells[(j,i-1)]
        # set top-most row to all blanks
        for x in range(self._width):
            self._cells[(x,0)] = None

        self._fire_board_event()

    def add_board_listener(self, listener):
        self._listeners.append(listener)

    def remove_board_listener(self, listener):
        self._listeners.remove(listener)

    def _fire_board_event(self):
        evt = events.BoardEvent(self)
        for listener in self._listeners:
            listener.board_changed(evt)

    def __str__(self):
        str_list = []
        for y in range(self._height):
            for x in range(self._width):
                if self._cells[(x,y)] == None:
                    str_list.append('. ')
                else:
                    str_list.append(self._cells[(x,y)] + ' ')
            if y < self._height - 1:
                str_list.append('\n')
        return ''.join(str_list)


# UNIT TESTING
if __name__ == '__main__':
    b = Board(width=10, height=20)
    lpiece = Piece(b, Piece.types['L'], (5, 1))
    rpiece = Piece.get_random_piece(b)
    rpiece.center = (5, 5)
    print 'lpiece:', lpiece
    print 'rpiece:', rpiece

    b.add_piece(lpiece, False)
    b.add_piece(rpiece, True)
    print b

    lpiece.move('RIGHT')
    print 'moved lpiece RIGHT.'
    lpiece.move('ROTATE')
    print 'ROTATEd lpiece.'
    counter = 0
    for x in range(20):
        if not rpiece.move('DOWN'):
            break
        counter += 1
    print "moved rpiece DOWN x%d." % counter
    lpiece.move('FALL')
    print 'moved lpiece to FALL.'
    print lpiece
    print rpiece
    print b

    print
    print 'removing last row ...'
    print 'removing lpiece ...'
    b.remove_piece(lpiece)
    b.remove_row(19)
    print b


